use std::rc::{Rc, Weak};
use std::cell::RefCell;
use std::collections::HashMap;
use r9_macro::SharedFromSelf;
use r9_macro_derive::SharedFromSelf;
use r9_macro_derive::Singleton;
use actix::{Actor, StreamHandler};
use actix_web_actors::ws::{self, Message};
use actix_web::{
    web,
    App as ActixApp,
    HttpRequest,
    HttpResponse,
    HttpServer,
    Error
    };
use std::sync::{Arc};
use tokio::sync::Mutex;
use std::thread;
use std::thread::JoinHandle;
use std::time::Duration;
use actix::prelude::*;
use bytes::{BufMut, BytesMut};
use actix_web::web::Bytes;
use futures::FutureExt;
use tokio::io::{AsyncReadExt, Interest};
use crate::app::UserApp;
use crate::{MAX_PACKET_LEN};
use tokio::net::TcpStream;
use tokio::runtime::Runtime;
use crate::app::user_app::DownStreamPack;
use tokio::io::{AsyncWriteExt, Ready};

struct AppStateWithCounter {
    counter: std::sync::Mutex<i32>, // <- Mutex is necessary to mutate safely across threadsds
}

#[derive(SharedFromSelf)]
#[derive(Singleton)]
pub struct WsListener {
    _self_wp: Weak::<RefCell::<Self>>,
    work_thread: Option<JoinHandle<()>>,
    socket_hash: HashMap<u16, Recipient<DownStreamMessage>>,
    cur_socket_handle: u16,
    pub upstream: std::sync::Weak<Mutex<TcpStream>>,
    tokio_rt: Runtime,
    down_pack_queue: Arc<std::sync::Mutex<r9::Queue<DownStreamPack>>>,
    up_pack_queue: Arc<std::sync::Mutex<r9::Queue<DownStreamPack>>>,
}

#[derive(Message)]
#[rtype(result = "()")]
pub struct DownStreamMessage {
    data: actix_web::web::Bytes,
    //data: BytesMut,
}

/// Define HTTP actor
pub struct WsConn {
    socket_handle: u16,
    recv_buf_len: usize,
    recv_buf: BytesMut,
}

impl Actor for WsConn {
    type Context = ws::WebsocketContext<Self>;
}

impl WsConn {
    fn parse_pkt(&mut self, bin: &actix_web::web::Bytes, ctx: &mut <WsConn as Actor>::Context) {
        let buf_len = bin.len();
        if buf_len > 0 {
            let mut already_read_bytes: usize = 0;
            {
                for i in 0..1024 * 64 * 2 {
                    self.recv_buf.put_i8(0);
                }
            }
            while true {
                let mut read_bytes = buf_len - already_read_bytes;
                if read_bytes > MAX_PACKET_LEN - self.recv_buf_len {
                    read_bytes = MAX_PACKET_LEN - self.recv_buf_len;
                }
                if read_bytes > 0 {
                    self.recv_buf[self.recv_buf_len..bin.len()].copy_from_slice(&bin);
                    self.recv_buf_len += read_bytes;
                    already_read_bytes += read_bytes;
                }

                let mut offset = 0;
                let mut pre_offect = 0;
                while true {
                    pre_offect = offset;
                    {
                        let pack_len = (self.recv_buf[0] as u16)+ ((self.recv_buf[1] as u16) << 8);
                        let msg_id = (self.recv_buf[2] as u16)+ ((self.recv_buf[3] as u16) << 8);
                        let seq_id = (self.recv_buf[4] as i32)+ ((self.recv_buf[5] as i32) << 8) +
                            ((self.recv_buf[7] as i32) << 16)+ ((self.recv_buf[7] as i32) << 24);
                        let magic_code = (self.recv_buf[8] as u16)+ ((self.recv_buf[9] as u16) << 8);
                        let ext_len = (self.recv_buf[10] as u16)+ ((self.recv_buf[11] as u16) << 8);
                        let mut tmp_bytes = BytesMut::with_capacity((pack_len + 24) as usize);
                        {
                            tmp_bytes.put_u16_le(pack_len);
                            tmp_bytes.put_u16_le(msg_id);
                            tmp_bytes.put_i32_le(seq_id);
                            tmp_bytes.put_u16_le(magic_code);
                            tmp_bytes.put_u16_le(self.socket_handle); //tmp_bytes.put_u16_le(ext_len);
                            tmp_bytes.put_i32_le(0);
                            tmp_bytes.put_i64_le(0);
                            for i in 24..tmp_bytes.capacity() {
                                tmp_bytes.put_i8(0);
                            }
                            tmp_bytes[24 .. (24+ pack_len as usize)].copy_from_slice(&bin[12..(12 + pack_len as usize)]);
                            println!("recv buf_len:{0}", tmp_bytes.len());
                        }
                        {
                            UserApp::instance().borrow().add_down_stream_pack(self.socket_handle, tmp_bytes);
                        }
                        offset += (pack_len  as usize) + 12;
                    }
                    if pre_offect >= offset || offset >= self.recv_buf_len {
                        break;
                    }
                }//end while
                if offset > 0 && offset < self.recv_buf_len {
                    self.recv_buf[..0].copy_from_slice(&bin);
                }
                self.recv_buf_len -= offset;
                if self.recv_buf_len > 0 {
                    self.recv_buf_len = 0;
                }
                if already_read_bytes >= bin.len() {
                    already_read_bytes = 0;
                    break;
                }
            }
        }
    }

}

impl StreamHandler<Result<Message, ws::ProtocolError>> for WsConn {

    fn started(&mut self, ctx: &mut Self::Context) {
        let  a: Recipient<DownStreamMessage> = ctx.address().recipient();
        WsListener::instance().borrow_mut().on_connect(self, a);
    }
    fn handle(&mut self, item: Result<Message, ws::ProtocolError>, ctx: &mut Self::Context) {
        match item {
            Ok(Message::Text(text)) => {

            },
            Ok(Message::Ping(msg)) => {
                ctx.pong(&msg)
            },
            Ok(Message::Binary(bin)) => {
                self.parse_pkt(&bin, ctx);
            }
            Ok(Message::Close(reason)) => {
                ctx.close(reason);
                WsListener::instance().borrow_mut().on_disconnect(self.socket_handle);
            },
            _ => {},
        }
    }
}

impl Handler<DownStreamMessage> for WsConn {
    type Result = ();

    fn handle(&mut self, msg: DownStreamMessage, ctx: &mut Self::Context) -> Self::Result {
        ctx.binary(msg.data);
    }
}

async fn index(req: HttpRequest, stream: web::Payload) -> Result<HttpResponse, Error> {
    let mut conn = WsConn{
        socket_handle: 0,
        recv_buf_len: 0,
        recv_buf: BytesMut::with_capacity(1024 * 64 * 2)
    };
    conn.recv_buf.resize(conn.recv_buf.capacity(), 0);
    let resp = ws::start(conn, &req, stream);
    println!("{:?}", resp);
    resp
}

async fn upstream_enter(stream_arc: Arc<Mutex<TcpStream>>, down_pack_queue: Arc<std::sync::Mutex::<r9::Queue<DownStreamPack>>>,
                        up_pack_queue: Arc<std::sync::Mutex::<r9::Queue<DownStreamPack>>>) {
    tokio::spawn(async move {
        println!("upstream_enter2");
        let mut recv_buf = BytesMut::with_capacity(1024 * 64 * 3);
        let mut recv_buf_len = 0;
        {
            for i in 0..recv_buf.capacity() {
                recv_buf.put_i8(0);
            }
        }
        loop {
            tokio::time::sleep(Duration::from_millis(10)).await;
            let mut stream = stream_arc.lock().await;
            let ready = stream.ready(Interest::READABLE | Interest::WRITABLE).await;
            match ready {
                Ok(r) => {
                    {
                        if r.is_readable() {
                            // Try to read data, this may still fail with `WouldBlock`
                            // if the readiness event is a false positive.
                            let mut data = vec![0; 1024 * 64];
                            match stream.try_read(&mut data) {
                                Ok(len) => {
                                    println!("read {0} bytes",len);
                                    for i in 0..len {
                                        recv_buf[recv_buf_len + i] = data[i];
                                        //println!("sendmsg xxxxx idx:{0} val:{1} buf_val:{2} recv_buf_len:{3}", i, data[i], recv_buf[recv_buf_len + 1], recv_buf_len);
                                    }
                                    recv_buf_len += len;
                                    while recv_buf_len >= 16 {
                                        let pack_len = (recv_buf[0] as u16) + ((recv_buf[1] as u16) << 8);
                                        let msg_id = (recv_buf[2] as u16) + ((recv_buf[3] as u16) << 8);
                                        let seq_id = (recv_buf[4] as i32) + ((recv_buf[5] as i32) << 8) +
                                            ((recv_buf[7] as i32) << 16) + ((recv_buf[7] as i32) << 24);
                                        let magic_code = (recv_buf[8] as u16) + ((recv_buf[9] as u16) << 8);
                                        let rpc_error_code = (recv_buf[10] as u16) + ((recv_buf[11] as u16) << 8);
                                        let socket_handle = (recv_buf[12] as u16) + ((recv_buf[13] as u16) << 8);
                                        let ext_len = (recv_buf[14] as u16) + ((recv_buf[15] as u16) << 8);
                                        if ((pack_len + 16) as usize) > recv_buf_len {
                                            break;
                                        }
                                        let mut tmp_bytes = BytesMut::with_capacity((pack_len + 16) as usize);
                                        {
                                            tmp_bytes.put_u16_le(pack_len);
                                            tmp_bytes.put_u16_le(msg_id);
                                            tmp_bytes.put_i32_le(seq_id);
                                            tmp_bytes.put_u16_le(magic_code);
                                            tmp_bytes.put_u16_le(ext_len);
                                            for i in 0..pack_len {
                                                tmp_bytes.put_u8(recv_buf[(16 + i) as usize]);
                                            }
                                            let mut a_bytes = actix_web::web::Bytes::copy_from_slice(tmp_bytes.as_ref());
                                            {
                                                let node = Rc::new(RefCell::new(DownStreamPack{
                                                    holder: Default::default(),
                                                    socket_handle: socket_handle,
                                                    data: tmp_bytes,
                                                    entry: r9::ListHead::< crate::app::user_app::DownStreamPack>::new_node(Default::default()),
                                                }));
                                                node.borrow_mut().holder = Some(node.clone());
                                                node.borrow_mut().entry = r9::ListHead::< crate::app::user_app::DownStreamPack>::new_node(Rc::downgrade(&node));
                                                up_pack_queue.lock().unwrap().push(&node.borrow_mut().entry);
                                            }
                                            //a.do_send(msg);

                                            {
                                                println!("sendmsg socket_handle:{5} pack_len:{0} msg_id:{1} seq_id:{2} magic_code:{3} ext_len:{4}",
                                                         socket_handle,
                                                         pack_len,
                                                         msg_id,
                                                        seq_id,
                                                magic_code,
                                                ext_len);
                                            }
                                        }
                                        for i in (pack_len + 16) as usize..recv_buf_len {
                                            recv_buf[i - (pack_len + 16) as usize] = recv_buf[i];
                                        }
                                        recv_buf_len -= (pack_len + 16) as usize;
                                    }
                                }
                                Err(ref e) if e.kind() == std::io::ErrorKind::WouldBlock => {
                                    println!("read error 1");
                                    continue;
                                }
                                Err(e) => {
                                    println!("read error 2");
                                    //return Err(e.into());
                                }
                            }
                        }
                    }

                    {
                        if r.is_writable() {
                            let mut data = bytes::BytesMut::new();
                            {
                                down_pack_queue.lock().unwrap().fetch();
                                //down_pack_queue.get_mut().fetch();
                                let work_list = down_pack_queue.lock().unwrap().work_list.lock().unwrap().clone();
                                //let work_list = down_pack_queue.get_mut().work_list.lock().unwrap().clone();
                                while !work_list.borrow().empty() {
                                    let node = &work_list.borrow().first_entry();
                                    node.upgrade().unwrap().borrow_mut().entry.borrow_mut().del_init();
                                    data = node.upgrade().unwrap().borrow().data.clone();
                                    break;
                                    //stream.write_all(node.upgrade().unwrap().borrow().data.as_ref()).await;
                                    //stream.write_all(data.as_ref()).await;
                                }
                            }
                            if data.len() > 0 {
                                println!("write_all2 len:{0}", data.len());
                                stream.write_all(data.as_ref()).await;
                            }
                        }
                    }
                }
                Err(e) => {

                }
            }
        }
    });
}

impl WsListener {

    pub fn new() -> Self {
        let p = Self{
            cur_socket_handle: 0,
            _self_wp: Default::default(),
            work_thread: Default::default(),
            socket_hash: Default::default(),
            upstream: Default::default(),
            tokio_rt: tokio::runtime::Builder::new_multi_thread()
                .enable_all()
                .build()
                .unwrap(),
            down_pack_queue: r9::Queue::<DownStreamPack>::new_ex(),
            up_pack_queue: r9::Queue::<DownStreamPack>::new_ex(),
        };
        return p;
    }

    pub fn init(&mut self) {
        //self.work_thread = Arc::new(Some(1));
        {
            self.work_thread = Some(thread::spawn(|| {
                println!("hello2");
                let mut rt = tokio::runtime::Builder::new_multi_thread()
                    .enable_all()
                    .build()
                    .unwrap();
                rt.block_on(async {
                    let counter = web::Data::new(AppStateWithCounter {
                        counter: std::sync::Mutex::new(0),
                    });
                    println!("hello2333");
                    HttpServer::new(move || {
                        // move counter into the closure
                        ActixApp::new() // <- register the created data
                            .app_data(counter.clone())
                            .route("/", web::get().to(index))
                    })
                        .bind(("0.0.0.0", 8080))?
                        .run()
                        .await
                });
            }));
            {
                let upstream = self.upstream.clone();
                let down_pack_queue = self.down_pack_queue.clone();
                let up_pack_queue = self.up_pack_queue.clone();
                self.tokio_rt.spawn(async move {
                    let mut result = TcpStream::connect("192.168.100.39:7616").await;
                    match result {
                        Ok(v) => {
                            let a = Arc::new(Mutex::new(v));
                            WsListener::instance().borrow_mut().upstream = Arc::downgrade(&a);
                            println!("connect ok");
                            upstream_enter(a, down_pack_queue, up_pack_queue).await;
                        },
                        Err(e) => {
                            println!("connect err")
                        }
                    }
                });
            }
            println!("hello3");
        }
    }

    pub fn update(&self) {
        let mut data = bytes::BytesMut::new();
        let mut socket_handle = 0;
        {
            self.up_pack_queue.lock().unwrap().fetch();
            //down_pack_queue.get_mut().fetch();
            let work_list = self.up_pack_queue.lock().unwrap().work_list.lock().unwrap().clone();
            //let work_list = down_pack_queue.get_mut().work_list.lock().unwrap().clone();
            while !work_list.borrow().empty() {
                let node = &work_list.borrow().first_entry();
                node.upgrade().unwrap().borrow_mut().entry.borrow_mut().del_init();
                data = node.upgrade().unwrap().borrow().data.clone();
                socket_handle = node.upgrade().unwrap().borrow().socket_handle;
                break;
                //stream.write_all(node.upgrade().unwrap().borrow().data.as_ref()).await;
                //stream.write_all(data.as_ref()).await;
            }
        }
        if data.len() > 0 {
            println!("write_all3 socket_handle:{0} len:{1}", socket_handle, data.len());
            match self.socket_hash.get(&socket_handle){
                Some(v) => {
                    let msg = DownStreamMessage{
                        data: actix_web::web::Bytes::copy_from_slice(data.as_ref())
                    };
                    v.do_send(msg);
                },
                None => {

                }
            }
        }
    }

    pub fn uninit(&mut self) {

    }

    pub fn add_down_stream_pack(&self, data: Weak::<RefCell::<DownStreamPack>>) {
        println!("add_down_stream_pack2 len:{0}", data.upgrade().unwrap().borrow().data.len());
        let socket_handle = data.upgrade().unwrap().borrow().socket_handle;
        self.down_pack_queue.lock().as_ref().unwrap().push(&data.upgrade().unwrap().borrow().entry);
    }
    fn on_connect(&mut self, conn: &mut WsConn, socket: Recipient<DownStreamMessage>) {
        self.cur_socket_handle += 1;
        conn.socket_handle = self.cur_socket_handle;
        self.socket_hash.insert(conn.socket_handle, socket);
        println!("on_connect socket_handle:{0}", conn.socket_handle);
    }

    fn on_disconnect(&mut self, socket_handle: u16) {
        self.socket_hash.remove(&socket_handle);
        println!("on_disconnect socket_handle:{0}", socket_handle);
    }

}
